#include <v_lib.h>

void V_CALL deriv(vec_t dest[],vec_t x,vec_t y[])
{
	/*
	 * d^2y/dx^2 = -y <==>
	 * y = [u,v], where u - speed, v - position
	 * du/dx = -v
	 * dv/dx = u
	 * 
	 */
	dest[1] = y[0];
	dest[0] = - y[1];
}

vboolean V_CALL monitor(void *param,vec_t x,vec_t y[])
{
	/*
	printf("monitor: x = %g, y[0] = %g, y[1] = %g\n",x,y[0],y[1]);
	*/
	printf("%d\t%d\t%d\n",(int)(1000 * x),(int)(1000 * y[0]),(int)(1000 * y[1]));
	return true;
}

int main()
{
	vhandle_t solver;
	vec_t y[2];
	
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	
	solver = odeCreateSolver(2,deriv,true);

	if(!solver)
		return -1;
	
	y[0] = 0;
	y[1] = 1;

	// odeSetMonitor(solver,monitor,NULL,0.25,false);
	
	{
		u32_t start = timeTickCount();
		odeIntegrate(solver,y,0,100000 * M_2PI + M_PI2,1e-8,1e-6,1e-30);
		start = timeTickCount() - start;
		printf("done: %d ms,y[0]=%g,y[1]=%g\n",start,y[0],y[1]);
	}
	
	odeDestroySolver(solver);
	
	return 0;
}
