#include <../source/TAL/objbase.h>

/*
 * fftwrap.c
 * This module contains both platform-specific wrappers
 * and simple implementation of the Fast Fourier Transform.
 */

#define	power2(op)	(1<<(op))

#ifdef	CONFIG_HAVE_NSP

#include <../support/nsp/include/nsp.h>
#include <../support/nsp/include/nspfft.h>

void	V_GLOBAL
fftCalcNip(
	vec_t *ReOut,vec_t *ImOut,
	vec_t *ReIn,vec_t *ImIn,
	unsigned size_power,vec_t dir_sign)
{
	if(!ReOut || !ImOut || !ReIn || !ImIn || !size_power)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"fftCalcNip(): incorrect parameters.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	if(!ReOut || !ImOut || !ReIn || !ImIn || !size_power)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"fftCalcNip(): incorrect parameters.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	if(dir_sign > 0)	/* Forward FFT */
	{
#ifdef	CONFIG_DOUBLEVEC
		nspzrFftNip(ReIn,ImIn,ReOut,ImOut,size_power,NSP_Forw);
#else	/* !CONFIG_DOUBLEVEC */
		nspcrFftNip(ReIn,ImIn,ReOut,ImOut,size_power,NSP_Forw);
#endif	/* !CONFIG_DOUBLEVEC */
	}
	else				/* Inverse FFT */
	{
#ifdef	CONFIG_DOUBLEVEC
		nspzrFftNip(ReIn,ImIn,ReOut,ImOut,size_power,NSP_Inv);
#else	/* !CONFIG_DOUBLEVEC */
		nspcrFftNip(ReIn,ImIn,ReOut,ImOut,size_power,NSP_Inv);
#endif	/* !CONFIG_DOUBLEVEC */
	}
}

void	V_GLOBAL
fftRealNip(vec_t *dest,vec_t *src,unsigned size_power,vboolean forward)
{
	if(!dest||!src||!size_power)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"fftRealNip(): incorrect parameters.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	if(forward)	/* Forward FFT */
	{
#ifdef	CONFIG_DOUBLEVEC
		nspdRealFftlNip(src,dest,size_power,NSP_Forw|NSP_OutRCPack);
#else	/* !CONFIG_DOUBLEVEC */
		nspsRealFftlNip(src,dest,size_power,NSP_Forw|NSP_OutRCPack);
#endif	/* !CONFIG_DOUBLEVEC */
	}
	else				/* Inverse FFT */
	{
#ifdef	CONFIG_DOUBLEVEC
		nspdCcsFftlNip(src,dest,size_power,NSP_Inv|NSP_InRCPack);
#else	/* !CONFIG_DOUBLEVEC */
		nspsCcsFftlNip(src,dest,size_power,NSP_Inv|NSP_InRCPack);
#endif	/* !CONFIG_DOUBLEVEC */
	}
}	

#elif	defined(CONFIG_HAVE_GSL) /* !CONFIG_HAVE_NSP */

#include <gsl_errno.h>
#include <gsl_fft_complex.h>

void	V_GLOBAL
fftCalcNip(
	vec_t *ReOut,vec_t *ImOut,
	vec_t *ReIn,vec_t *ImIn,
	unsigned size_power,vec_t dir_sign)
{
	size_t i,j,len;
	double *packed;
	if(!ReOut || !ImOut || !ReIn || !ImIn || !size_power)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"fftCalcNip(): incorrect parameters.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	len=power2(size_power);
	packed=(double*)memAlloc((len << 1) * sizeof(double));
	for(i=j=0;i<len;i++,j+=2)
	{
		packed[j]=ReIn[i];
		packed[j+1]=ImIn[i];
	}
	if(dir_sign > 0)	/* Forward FFT */
		gsl_fft_complex_radix2_forward(packed,1,len);
	else				/* Inverse FFT */
		gsl_fft_complex_radix2_inverse(packed,1,len);
	for(i=j=0;i<len;i++,j+=2)
	{
		ReOut[i]=packed[j];
		ImOut[i]=packed[j+1];
	}
	memFree(packed);
}

#else	/* !CONFIG_HAVE_GSL */

/*
 * Cut - and - paste from 'Numerical recipies in C'.
 */
#define NRC_FFT_SWAP(a, b) tempr = (a); (a) = (b); (b) = tempr

/*
 * Replaces data by its discrete Fourier transform, if isign is input as 1;
 * or replaces data by nn times its inverse discrete Fourier transform,
 * if isign is input as -1.
 * data is a complex array of length nn, input as a real array data[1..2*nn].
 * nn MUST be an integer power of 2.
*/

V_LOCAL void
nrc_four1(vec_t data[], int nn, int isign)
{
	int i, n, mmax, m, j, istep;
	vec_t wtemp, wr, wpr, wpi, wi, theta;
	vec_t tempr, tempi;
	n = nn << 1;
	j = 1;
	/* This is the bit-reversal section of the routine. */
	for (i = 1; i < n; i += 2)
	{
    	if (j > i)
		{
			/* Exchange the two complex numbers. */
	  		NRC_FFT_SWAP(data[j  ],data[i  ]);
			NRC_FFT_SWAP(data[j+1],data[i+1]);
		};
		m = n >> 1;
		while (m >= 2 && j > m)
		{
			j  -= m;
			m >>= 1;
		};
		j += m;
	};
	/* Here begins the Danielson-Lanczos section of the routine. */
	mmax = 2;
	while (n > mmax)
	{
		/* Initialize trigonometric recurrence. */
		theta	= (vec_t)6.28318530717959/(isign*mmax);
		wtemp	= (vec_t)sin(0.5*theta);
		wpr	= (vec_t)(-2.0)*wtemp*wtemp;
		wpi	= (vec_t)sin(theta);
		wr	= 1;
		wi	= 0;
		/* Here are the two nested inner loops. */
		istep	= 2*mmax;
		for (m = 1; m < mmax; m += 2)
		{
			for (i = m; i <= n; i += istep)
			{
				/* This is the Danielson-Lanczos formula. */
				j = i + mmax;
				tempr = wr*data[j  ] - wi*data[j+1];
				tempi = wr*data[j+1] + wi*data[j  ];
				data[j  ] = data[i  ] - tempr;
				data[j+1] = data[i+1] - tempi;
				data[i  ] += tempr;
				data[i+1] += tempi;
			}; /* Trigonometric recurrence. */
			wr = (wtemp = wr)*wpr - wi*wpi + wr;
			wi = wi*wpr + wtemp*wpi + wi;
		};
		mmax = istep;
	};
}

/*
 * Again, cut-and-paste
 */

/*
  Calculates the Fourier Transform of a set of 2n real-valued data points
  stored in array data[1..2n] replacing it by the positive frequency half
  of it's complex Fourier Transform if isign is set to +1.
  The real-valued first and last components of the complex transform
  are returned as elements data[1] and data[2] respectively.
  n must be a power of 2.
  Calculates the inverse transform of a complex data array
  which is the transform of real data if isign is set to -1.
  But the inverse transform array should be multiplied by 1/n.
*/

V_LOCAL void
nrc_realft(vec_t data[],int n,int isign)
{
	int	 i, i1, i2, i3, i4, n2p3;
	vec_t c1 = 0.5, c2, h1r, h1i, h2r, h2i;
	vec_t wr, wi, wpr, wpi, wtemp, theta;
	theta = (vec_t)M_PI/(vec_t)n;
	if(isign == 1)
	{
		c2 = -(vec_t)0.5;
		nrc_four1(data, n, 1);
	}
	else
	{
		c2 = (vec_t)0.5;
		theta = -theta;
	}
	wtemp = (vec_t)sin(0.5*theta);
	wpr = -(vec_t)2.0 * wtemp * wtemp;
	wpi = (vec_t)sin(theta);
	wr = 1 + wpr;
	wi = wpi;
	n2p3 = 2*n + 3;
	for (i = 2; i <= n/2; i++ )
	{
		i4	 =   1 + (i3 = n2p3 - (i2 = 1 + (i1 = i + i - 1)));
		h1r	 =  c1*(data[i1] + data[i3]);
		h1i	 =  c1*(data[i2] - data[i4]);
		h2r	 = -c2*(data[i2] + data[i4]);
		h2i	 =  c2*(data[i1] - data[i3]);
		data[i1]	 =  h1r + wr*h2r - wi*h2i;
		data[i2]	 =  h1i + wr*h2i + wi*h2r;
		data[i3]	 =  h1r - wr*h2r + wi*h2i;
		data[i4]	 = -h1i + wr*h2i + wi*h2r;
		wr	 = (wtemp = wr) * wpr - wi * wpi + wr;
		wi	 =  wi*wpr + wtemp*wpi + wi;
	}
	if (isign == 1)
    {
		data[1]	 = (h1r = data[1]) + data[2];
		data[2]	 =  h1r - data[2];
	}
	else
    {
		data[1]	 =  c1*((h1r = data[1]) + data[2]);
		data[2]	 =  c1*(h1r - data[2]);
		nrc_four1(data, n, -1);
    }
}

/*
 * Calculates the cosine transform of a set y[1..n] of real-valued data points.
 * The transformed data replaces the original data in array y.
 * n must be a power of 2.
 * Set isign to +1 for a transform and to -1 for an inverse transform.
 * But the inverse transform array should be multiplied by 2/n.
 */
V_LOCAL void
nrc_cosft (vec_t y[],unsigned n,unsigned isign)
{
    unsigned j, m, n2;
    vec_t enf0, even, odd, sum, sume, sumo, y1, y2;
    vec_t theta, wi = 0.0, wr = 1.0, wpi, wpr, wtemp;
  
    theta = M_PI / (vec_t) n;
    wtemp = (vec_t) sin (0.5 * theta);
    wpr = - (vec_t)2.0 * wtemp * wtemp;
    wpi = (vec_t) sin (theta);
    sum = y[1];
    m = n >> 1;
    n2 = n + 2;
    for (j = 2; j <= m; j++)
    {
        wr = (wtemp = wr) * wpr - wi * wpi + wr;
        wi = wi * wpr + wtemp * wpi + wi;
        y1 = (vec_t)0.5 * (y[j] + y[n2 - j]);
        y2 = (y[j] - y[n2 - j]);
        y[j] = y1 - wi * y2;
        y[n2 - j] = y1 + wi * y2;
        sum += wr * y2;
    }
    nrc_realft (y, m, 1);
    y[2] = sum;
    for (j = 4; j <= n; j += 2)
    {
        sum += y[j];
        y[j] = sum;
    }
    if (isign == -1)
    {
        even = y[1];
        odd = y[2];
        for (j = 3; j <= n - 1; j += 2)
        {
            even += y[j];
            odd += y[j + 1];
        };
        enf0 = (vec_t)2.0 * (even - odd);
        sumo = y[1] - enf0;
        sume = ((vec_t)2 * odd / n) - sumo;
        y[1] = (vec_t)0.5 * enf0;
        y[2] -= sume;
        for (j = 3; j <= n - 1; j += 2)
        {
            y[j] -= sumo;
            y[j + 1] -= sume;
        }
    }
}

/*
 * Wrappers
 */

void	V_GLOBAL
fftCalcNip(
	vec_t *ReOut,vec_t *ImOut,
	vec_t *ReIn,vec_t *ImIn,
	unsigned size_power,vec_t dir_sign)
{
	unsigned i,len;
	vec_t *packed, scale;
	if(!ReOut || !ImOut || !ReIn || !ImIn || !size_power)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"fftCalcNip(): incorrect parameters.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	len = power2(size_power);
	packed=(vec_t*)memAlloc((len << 1) * sizeof(vec_t));
	for(i=0;i<len;i++)
	{
		packed[i<<1]=ReIn[i];
		packed[(i<<1)+1]=ImIn[i];
	}
	if(dir_sign > 0)
	{
		nrc_four1(packed-1,len,1);
		for(i=0;i<len;i++)
		{
			ReOut[i] = packed[i<<1];
			ImOut[i] = packed[(i<<1)+1];
		}
	}
	else
	{
		nrc_four1(packed-1,len,-1);
		scale = (vec_t)1.0 / (vec_t)len;
		for(i=0;i<len;i++)
		{
			ReOut[i]=packed[i<<1] * scale;
			ImOut[i]=packed[(i<<1)+1] * scale;
		}
	}
	memFree(packed);
}

void	V_GLOBAL
fftRealNip(vec_t *dest,vec_t *src,unsigned size_power,vboolean forward)
{
	vec_t scale;
	unsigned i, len;
	if(!dest||!src||!size_power)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"fftRealNip(): incorrect parameters.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	len = power2(size_power);
	memCopy(dest,src,len * sizeof(vec_t));
	nrc_realft(dest-1,len >> 1,forward ? 1 : -1);
	if(!forward)
	{
		scale = (vec_t)2.0 / (vec_t)len;
		for(i=0;i<len;i++)
			dest[i] *= scale;
	}
}	

void	V_GLOBAL
cosftNip(vec_t *dest,vec_t *src,unsigned size_power,vboolean forward)
{
	vec_t scale;
	unsigned i, len;
	if(!dest||!src||!size_power)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"cosftNip(): incorrect parameters.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	len = power2(size_power);
	memCopy(dest,src,len * sizeof(vec_t));
	nrc_cosft(dest - 1,len,forward ? 1 : -1);
	if(!forward)
	{
		scale = (vec_t)2.0 / (vec_t)len;
		for(i=0;i<len;i++)
			dest[i] *= scale;
	}
}	

#endif	/* !CONFIG_HAVE_NSP */
